@@ -14,13 +14,16 @@
  * limitations under the License.
  */
 
+#include <bson/bson-prelude.h>
+
 
 #ifndef BSON_ISO8601_PRIVATE_H
 #define BSON_ISO8601_PRIVATE_H
 
 
-#include "bson-compat.h"
-#include "bson-macros.h"
+#include <bson/bson-compat.h>
+#include <bson/bson-macros.h>
+#include <bson/bson-string.h>
 
 
 BSON_BEGIN_DECLS
@@ -31,6 +34,16 @@ _bson_iso8601_date_parse (const char *str,
                           int64_t *out,
                           bson_error_t *error);
 
+/**
+ * _bson_iso8601_date_format:
+ * @msecs_since_epoch: A positive number of milliseconds since Jan 1, 1970.
+ * @str: The string to append the ISO8601-formatted to.
+ *
+ * Appends a date formatted like "2012-12-24T12:15:30.500Z" to @str.
+ */
+void
+_bson_iso8601_date_format (int64_t msecs_since_epoch, bson_string_t *str);
+
 BSON_END_DECLS
 
 
