@@ -18,9 +18,9 @@
 #include <bson/bson.h>
 #include <stdarg.h>
 
-#include "mongoc/mongoc-error.h"
-#include "mongoc/mongoc-buffer-private.h"
-#include "mongoc/mongoc-trace-private.h"
+#include "mongoc-error.h"
+#include "mongoc-buffer-private.h"
+#include "mongoc-trace-private.h"
 
 
 #undef MONGOC_LOG_DOMAIN
@@ -31,8 +31,15 @@
 #endif
 
 
-#define SPACE_FOR(_b, _sz) \
-   (((ssize_t) (_b)->datalen - (ssize_t) (_b)->len) >= (ssize_t) (_sz))
+static void
+make_space_for (mongoc_buffer_t *buffer, size_t data_size)
+{
+   if (buffer->len + data_size > buffer->datalen) {
+      buffer->datalen = bson_next_power_of_two (buffer->len + data_size);
+      buffer->data = (uint8_t *) buffer->realloc_func (
+         buffer->data, buffer->datalen, buffer->realloc_data);
+   }
+}
 
 
 /**
@@ -55,7 +62,7 @@ _mongoc_buffer_init (mongoc_buffer_t *buffer,
                      bson_realloc_func realloc_func,
                      void *realloc_data)
 {
-   BSON_ASSERT (buffer);
+   BSON_ASSERT_PARAM (buffer);
    BSON_ASSERT (buflen || !buf);
 
    if (!realloc_func) {
@@ -89,7 +96,7 @@ _mongoc_buffer_init (mongoc_buffer_t *buffer,
 void
 _mongoc_buffer_destroy (mongoc_buffer_t *buffer)
 {
-   BSON_ASSERT (buffer);
+   BSON_ASSERT_PARAM (buffer);
 
    if (buffer->data && buffer->realloc_func) {
       buffer->realloc_func (buffer->data, 0, buffer->realloc_data);
@@ -111,7 +118,7 @@ _mongoc_buffer_destroy (mongoc_buffer_t *buffer)
 void
 _mongoc_buffer_clear (mongoc_buffer_t *buffer, bool zero)
 {
-   BSON_ASSERT (buffer);
+   BSON_ASSERT_PARAM (buffer);
 
    if (zero) {
       memset (buffer->data, 0, buffer->datalen);
@@ -130,23 +137,12 @@ _mongoc_buffer_append (mongoc_buffer_t *buffer,
 
    ENTRY;
 
-   BSON_ASSERT (buffer);
+   BSON_ASSERT_PARAM (buffer);
    BSON_ASSERT (data_size);
 
    BSON_ASSERT (buffer->datalen);
-   BSON_ASSERT ((buffer->datalen + data_size) < INT_MAX);
 
-   if (!SPACE_FOR (buffer, data_size)) {
-      if (buffer->len) {
-         memmove (&buffer->data[0], buffer->data, buffer->len);
-      }
-
-      if (!SPACE_FOR (buffer, data_size)) {
-         buffer->datalen = bson_next_power_of_two (data_size + buffer->len);
-         buffer->data = (uint8_t *) buffer->realloc_func (
-            buffer->data, buffer->datalen, NULL);
-      }
-   }
+   make_space_for (buffer, data_size);
 
    buf = &buffer->data[buffer->len];
 
@@ -178,7 +174,7 @@ bool
 _mongoc_buffer_append_from_stream (mongoc_buffer_t *buffer,
                                    mongoc_stream_t *stream,
                                    size_t size,
-                                   int32_t timeout_msec,
+                                   int64_t timeout_msec,
                                    bson_error_t *error)
 {
    uint8_t *buf;
@@ -186,41 +182,40 @@ _mongoc_buffer_append_from_stream (mongoc_buffer_t *buffer,
 
    ENTRY;
 
-   BSON_ASSERT (buffer);
-   BSON_ASSERT (stream);
+   BSON_ASSERT_PARAM (buffer);
+   BSON_ASSERT_PARAM (stream);
    BSON_ASSERT (size);
 
    BSON_ASSERT (buffer->datalen);
-   BSON_ASSERT ((buffer->datalen + size) < INT_MAX);
 
-   if (!SPACE_FOR (buffer, size)) {
-      if (buffer->len) {
-         memmove (&buffer->data[0], buffer->data, buffer->len);
-      }
-
-      if (!SPACE_FOR (buffer, size)) {
-         buffer->datalen = bson_next_power_of_two (size + buffer->len);
-         buffer->data = (uint8_t *) buffer->realloc_func (
-            buffer->data, buffer->datalen, NULL);
-      }
-   }
+   make_space_for (buffer, size);
 
    buf = &buffer->data[buffer->len];
 
    BSON_ASSERT ((buffer->len + size) <= buffer->datalen);
 
-   ret = mongoc_stream_read (stream, buf, size, size, timeout_msec);
-   if (ret != size) {
+   if (BSON_UNLIKELY (!bson_in_range_signed (int32_t, timeout_msec))) {
+      // CDRIVER-4589
       bson_set_error (error,
                       MONGOC_ERROR_STREAM,
                       MONGOC_ERROR_STREAM_SOCKET,
-                      "Failed to read %" PRIu64
-                      " bytes: socket error or timeout",
-                      (uint64_t) size);
+                      "timeout_msec value %" PRId64
+                      " exceeds supported 32-bit range",
+                      timeout_msec);
       RETURN (false);
    }
 
-   buffer->len += ret;
+   ret = mongoc_stream_read (stream, buf, size, size, (int32_t) timeout_msec);
+   if (bson_cmp_not_equal_su (ret, size)) {
+      bson_set_error (error,
+                      MONGOC_ERROR_STREAM,
+                      MONGOC_ERROR_STREAM_SOCKET,
+                      "Failed to read %zu bytes: socket error or timeout",
+                      size);
+      RETURN (false);
+   }
+
+   buffer->len += (size_t) ret;
 
    RETURN (true);
 }
@@ -241,7 +236,7 @@ ssize_t
 _mongoc_buffer_fill (mongoc_buffer_t *buffer,
                      mongoc_stream_t *stream,
                      size_t min_bytes,
-                     int32_t timeout_msec,
+                     int64_t timeout_msec,
                      bson_error_t *error)
 {
    ssize_t ret;
@@ -249,55 +244,63 @@ _mongoc_buffer_fill (mongoc_buffer_t *buffer,
 
    ENTRY;
 
-   BSON_ASSERT (buffer);
-   BSON_ASSERT (stream);
+   BSON_ASSERT_PARAM (buffer);
+   BSON_ASSERT_PARAM (stream);
 
    BSON_ASSERT (buffer->data);
    BSON_ASSERT (buffer->datalen);
 
    if (min_bytes <= buffer->len) {
-      RETURN (buffer->len);
+      BSON_ASSERT (bson_in_range_unsigned (ssize_t, buffer->len));
+      RETURN ((ssize_t) buffer->len);
    }
 
    min_bytes -= buffer->len;
 
-   if (buffer->len) {
-      memmove (&buffer->data[0], buffer->data, buffer->len);
-   }
-
-   if (!SPACE_FOR (buffer, min_bytes)) {
-      buffer->datalen = bson_next_power_of_two (buffer->len + min_bytes);
-      buffer->data = (uint8_t *) buffer->realloc_func (
-         buffer->data, buffer->datalen, buffer->realloc_data);
-   }
+   make_space_for (buffer, min_bytes);
 
    avail_bytes = buffer->datalen - buffer->len;
 
-   ret = mongoc_stream_read (
-      stream, &buffer->data[buffer->len], avail_bytes, min_bytes, timeout_msec);
-
-   if (ret == -1) {
+   if (BSON_UNLIKELY (!bson_in_range_signed (int32_t, timeout_msec))) {
+      // CDRIVER-4589
       bson_set_error (error,
                       MONGOC_ERROR_STREAM,
                       MONGOC_ERROR_STREAM_SOCKET,
-                      "Failed to buffer %u bytes",
-                      (unsigned) min_bytes);
+                      "timeout_msec value %" PRId64
+                      " exceeds supported 32-bit range",
+                      timeout_msec);
+      RETURN (false);
+   }
+
+   ret = mongoc_stream_read (stream,
+                             &buffer->data[buffer->len],
+                             avail_bytes,
+                             min_bytes,
+                             (int32_t) timeout_msec);
+
+   if (ret < 0) {
+      bson_set_error (error,
+                      MONGOC_ERROR_STREAM,
+                      MONGOC_ERROR_STREAM_SOCKET,
+                      "Failed to buffer %zu bytes",
+                      min_bytes);
       RETURN (-1);
    }
 
-   buffer->len += ret;
+   buffer->len += (size_t) ret;
 
    if (buffer->len < min_bytes) {
       bson_set_error (error,
                       MONGOC_ERROR_STREAM,
                       MONGOC_ERROR_STREAM_SOCKET,
-                      "Could only buffer %u of %u bytes",
-                      (unsigned) buffer->len,
-                      (unsigned) min_bytes);
+                      "Could only buffer %zu of %zu bytes",
+                      buffer->len,
+                      min_bytes);
       RETURN (-1);
    }
 
-   RETURN (buffer->len);
+   BSON_ASSERT (bson_in_range_unsigned (ssize_t, buffer->len));
+   RETURN ((ssize_t) buffer->len);
 }
 
 
@@ -318,34 +321,37 @@ ssize_t
 _mongoc_buffer_try_append_from_stream (mongoc_buffer_t *buffer,
                                        mongoc_stream_t *stream,
                                        size_t size,
-                                       int32_t timeout_msec)
+                                       int64_t timeout_msec)
 {
    uint8_t *buf;
    ssize_t ret;
 
    ENTRY;
 
-   BSON_ASSERT (buffer);
-   BSON_ASSERT (stream);
+   BSON_ASSERT_PARAM (buffer);
+   BSON_ASSERT_PARAM (stream);
    BSON_ASSERT (size);
 
    BSON_ASSERT (buffer->datalen);
-   BSON_ASSERT ((buffer->datalen + size) < INT_MAX);
 
-   if (!SPACE_FOR (buffer, size)) {
-      buffer->datalen = bson_next_power_of_two (size + buffer->len);
-      buffer->data =
-         (uint8_t *) buffer->realloc_func (buffer->data, buffer->datalen, NULL);
-   }
+   make_space_for (buffer, size);
 
    buf = &buffer->data[buffer->len];
 
    BSON_ASSERT ((buffer->len + size) <= buffer->datalen);
 
-   ret = mongoc_stream_read (stream, buf, size, 0, timeout_msec);
+   if (BSON_UNLIKELY (!bson_in_range_signed (int32_t, timeout_msec))) {
+      // CDRIVER-4589
+      MONGOC_ERROR ("timeout_msec value %" PRId64
+                    " exceeds supported 32-bit range",
+                    timeout_msec);
+      RETURN (-1);
+   }
+
+   ret = mongoc_stream_read (stream, buf, size, 0, (int32_t) timeout_msec);
 
    if (ret > 0) {
-      buffer->len += ret;
+      buffer->len += (size_t) ret;
    }
 
    RETURN (ret);
