@@ -25,6 +25,8 @@
 #include "mongoc-read-prefs-private.h"
 #include "mongoc-error-private.h"
 
+#include <bson-dsl.h>
+
 #define WITH_TXN_TIMEOUT_MS (120 * 1000)
 
 static void
@@ -257,8 +259,8 @@ retry:
        * actually apply the error label due to reply being NULL */
       _mongoc_client_session_unpin (session);
       if (reply) {
-         bson_copy_to_excluding_noinit (
-            &reply_local, reply, "errorLabels", NULL);
+         bsonBuildAppend (*reply,
+                          insert (reply_local, not(key ("errorLabels"))));
          _mongoc_error_copy_labels_and_upsert (
             &reply_local, reply, UNKNOWN_COMMIT_RESULT);
       }
@@ -1102,8 +1104,12 @@ mongoc_client_session_start_transaction (mongoc_client_session_t *session,
    BSON_ASSERT (session);
 
    ret = true;
-   server_stream = mongoc_cluster_stream_for_writes (
-      &session->client->cluster, session, NULL /* reply */, error);
+   server_stream =
+      mongoc_cluster_stream_for_writes (&session->client->cluster,
+                                        session,
+                                        NULL /* deprioritized servers */,
+                                        NULL /* reply */,
+                                        error);
    if (!server_stream) {
       ret = false;
       GOTO (done);
@@ -1252,13 +1258,14 @@ mongoc_client_session_commit_transaction (mongoc_client_session_t *session,
 
    /* For testing only, mock out certain kinds of errors. */
    if (session->fail_commit_label) {
-      bson_t labels;
+      bson_array_builder_t *labels;
 
       BSON_ASSERT (reply);
 
       bson_init (reply);
-      BSON_APPEND_ARRAY_BEGIN (reply, "errorLabels", &labels);
-      BSON_APPEND_UTF8 (&labels, "0", session->fail_commit_label);
+      BSON_APPEND_ARRAY_BUILDER_BEGIN (reply, "errorLabels", &labels);
+      bson_array_builder_append_utf8 (labels, session->fail_commit_label, -1);
+      bson_append_array_builder_end (reply, labels);
 
       /* Waste the test timeout, if there is one set. */
       if (session->with_txn_timeout_ms) {
@@ -1380,6 +1387,7 @@ _mongoc_client_session_from_iter (mongoc_client_t *client,
                                   bson_error_t *error)
 {
    ENTRY;
+   BSON_ASSERT_PARAM (client);
 
    /* must be int64 that fits in uint32 */
    if (!BSON_ITER_HOLDS_INT64 (iter) || bson_iter_int64 (iter) > 0xffffffff) {
