@@ -19,6 +19,12 @@
 #include <sys/syscall.h>
 #elif defined(_WIN32)
 #include <process.h>
+#elif defined(__APPLE__)
+#include <pthread.h>
+#elif defined(__FreeBSD__)
+#include <sys/thr.h>
+#elif defined(__NetBSD__)
+#include <lwp.h>
 #else
 #include <unistd.h>
 #endif
@@ -30,32 +36,50 @@
 #include "mongoc-thread-private.h"
 
 
-static mongoc_once_t once = MONGOC_ONCE_INIT;
-static mongoc_mutex_t gLogMutex;
+static bson_once_t once = BSON_ONCE_INIT;
+static bson_mutex_t gLogMutex;
 static mongoc_log_func_t gLogFunc = mongoc_log_default_handler;
-#ifdef MONGOC_TRACE
-static bool gLogTrace = true;
-#endif
+static bool gLogTrace = MONGOC_TRACE_ENABLED;
 static void *gLogData;
 
-static MONGOC_ONCE_FUN (_mongoc_ensure_mutex_once)
+static BSON_ONCE_FUN (_mongoc_ensure_mutex_once)
 {
-   mongoc_mutex_init (&gLogMutex);
+   bson_mutex_init (&gLogMutex);
 
-   MONGOC_ONCE_RETURN;
+   BSON_ONCE_RETURN;
 }
 
 void
 mongoc_log_set_handler (mongoc_log_func_t log_func, void *user_data)
 {
-   mongoc_once (&once, &_mongoc_ensure_mutex_once);
+   bson_once (&once, &_mongoc_ensure_mutex_once);
 
-   mongoc_mutex_lock (&gLogMutex);
+   bson_mutex_lock (&gLogMutex);
    gLogFunc = log_func;
    gLogData = user_data;
-   mongoc_mutex_unlock (&gLogMutex);
+   bson_mutex_unlock (&gLogMutex);
 }
 
+bool
+_mongoc_log_trace_is_enabled (void)
+{
+   return gLogTrace && MONGOC_TRACE_ENABLED;
+}
+
+void
+mongoc_log_trace_enable (void)
+{
+   /* Enable trace logging if-and-only-if tracing is enabled at configure-time,
+    * otherwise tracing remains disabled.
+    */
+   gLogTrace = MONGOC_TRACE_ENABLED;
+}
+
+void
+mongoc_log_trace_disable (void)
+{
+   gLogTrace = false;
+}
 
 /* just for testing */
 void
@@ -76,13 +100,11 @@ mongoc_log (mongoc_log_level_t log_level,
    char *message;
    int stop_logging;
 
-   mongoc_once (&once, &_mongoc_ensure_mutex_once);
+   bson_once (&once, &_mongoc_ensure_mutex_once);
 
    stop_logging = !gLogFunc;
-#ifdef MONGOC_TRACE
-   stop_logging =
-      stop_logging || (log_level == MONGOC_LOG_LEVEL_TRACE && !gLogTrace);
-#endif
+   stop_logging = stop_logging || (log_level == MONGOC_LOG_LEVEL_TRACE &&
+                                   !_mongoc_log_trace_is_enabled ());
    if (stop_logging) {
       return;
    }
@@ -93,9 +115,9 @@ mongoc_log (mongoc_log_level_t log_level,
    message = bson_strdupv_printf (format, args);
    va_end (args);
 
-   mongoc_mutex_lock (&gLogMutex);
+   bson_mutex_lock (&gLogMutex);
    gLogFunc (log_level, log_domain, message, gLogData);
-   mongoc_mutex_unlock (&gLogMutex);
+   bson_mutex_unlock (&gLogMutex);
 
    bson_free (message);
 }
@@ -138,6 +160,8 @@ mongoc_log_default_handler (mongoc_log_level_t log_level,
    char nowstr[32];
    int pid;
 
+   BSON_UNUSED (user_data);
+
    bson_gettimeofday (&tv);
    t = tv.tv_sec;
 
@@ -171,66 +195,49 @@ mongoc_log_default_handler (mongoc_log_level_t log_level,
    pid = syscall (SYS_gettid);
 #elif defined(_WIN32)
    pid = (int) _getpid ();
+#elif defined(__FreeBSD__)
+   long tid;
+   thr_self (&tid);
+   pid = (int) tid;
+#elif defined(__OpenBSD__)
+   pid = (int) getthrid ();
+#elif defined(__NetBSD__)
+   pid = (int) _lwp_self ();
+#elif defined(__APPLE__)
+   uint64_t tid;
+   pthread_threadid_np (0, &tid);
+   pid = (int) tid;
 #else
    pid = (int) getpid ();
 #endif
 
    fprintf (stream,
-            "%s.%04ld: [%5d]: %8s: %12s: %s\n",
+            "%s.%04" PRId64 ": [%5d]: %8s: %12s: %s\n",
             nowstr,
-            tv.tv_usec / 1000L,
+            (int64_t) (tv.tv_usec / 1000),
             pid,
             mongoc_log_level_str (log_level),
             log_domain,
             message);
 }
 
-bool
-_mongoc_log_trace_is_enabled (void)
-{
-#ifdef MONGOC_TRACE
-   return gLogTrace;
-#else
-   return false;
-#endif
-}
-
-void
-mongoc_log_trace_enable (void)
-{
-#ifdef MONGOC_TRACE
-   gLogTrace = true;
-#endif
-}
-
-void
-mongoc_log_trace_disable (void)
-{
-#ifdef MONGOC_TRACE
-   gLogTrace = false;
-#endif
-}
-
 void
 mongoc_log_trace_bytes (const char *domain, const uint8_t *_b, size_t _l)
 {
-   bson_string_t *str, *astr;
-   int32_t _i;
-   uint8_t _v;
-
-#ifdef MONGOC_TRACE
-   if (!gLogTrace) {
+   if (!_mongoc_log_trace_is_enabled ()) {
       return;
    }
-#endif
 
-   str = bson_string_new (NULL);
-   astr = bson_string_new (NULL);
-   for (_i = 0; _i < _l; _i++) {
-      _v = *(_b + _i);
+   bson_string_t *const str = bson_string_new (NULL);
+   bson_string_t *const astr = bson_string_new (NULL);
 
-      if ((_i % 16) == 0) {
-         bson_string_append_printf (str, "%05x: ", _i);
+   size_t _i;
+   for (_i = 0u; _i < _l; _i++) {
+      const uint8_t _v = *(_b + _i);
+      const size_t rem = _i % 16u;
+
+      if (rem == 0u) {
+         bson_string_append_printf (str, "%05zx: ", _i);
       }
 
       bson_string_append_printf (str, " %02x", _v);
@@ -240,18 +247,18 @@ mongoc_log_trace_bytes (const char *domain, const uint8_t *_b, size_t _l)
          bson_string_append (astr, " .");
       }
 
-      if ((_i % 16) == 15) {
+      if (rem == 15u) {
          mongoc_log (
             MONGOC_LOG_LEVEL_TRACE, domain, "%s %s", str->str, astr->str);
          bson_string_truncate (str, 0);
          bson_string_truncate (astr, 0);
-      } else if ((_i % 16) == 7) {
+      } else if (rem == 7u) {
          bson_string_append (str, " ");
          bson_string_append (astr, " ");
       }
    }
 
-   if (_i != 16) {
+   if (_i != 16u) {
       mongoc_log (
          MONGOC_LOG_LEVEL_TRACE, domain, "%-56s %s", str->str, astr->str);
    }
@@ -273,11 +280,9 @@ mongoc_log_trace_iovec (const char *domain,
    size_t _l = 0;
    uint8_t _v;
 
-#ifdef MONGOC_TRACE
-   if (!gLogTrace) {
+   if (!_mongoc_log_trace_is_enabled ()) {
       return;
    }
-#endif
 
    for (_i = 0; _i < _iovcnt; _i++) {
       _l += _iov[_i].iov_len;
